; ArrowLISP Library
; Copyright (C) 1998-2004 Nils M Holm. All rights reserved.
; See the file LICENSE of the ArrowLISP distribution
; for conditions of use.

; ---purpose---
; Sort a list using the Quicksort algorithm.

; ---args---
; A - list to sort
; P - predicate defining the desired order

; ---keywords---
; QUICKSORT function, quick sort, ordering, sorting

; ---see-also---
; sort, isort, orderedp

; ---example---
; :l lib/lessp.l
; (quicksort '(#5 #9 #1 #7 #3 #0 #2 #4 #8 #6) lessp)

:c lib/append.l
:c lib/caar.l
;c lib/cadr.l
;c lib/cddr.l
:c lib/lognot.l
:c lib/orderedp.l

; ---code---
(defun quicksort (a p)
	(label

	; Has X at least 3 arguments?
	((cddrp (lambda (x)
		(cond ((null (cdr x)) *f*)
			((null (cddr x)) *f*)
			(t *t*))))

	; Find the limits of A. Return (high-limit . low-limit).
	(find-limits (lambda (a)
		(label ((limits (lambda (a lo hi)
			(cond ((null a) (cons lo hi))
				((p lo (car a)) (limits (cdr a) (car a) hi))
				((p (car a) hi) (limits (cdr a) lo (car a)))
				(t (limits (cdr a) lo hi))))))
		(limits a (car a) (car a)))))

	; Test median for being safe. A median is safe, if it is not
	; equal to any of the limits of A.
	(safe-median-p (lambda (x limits)
		(cond ((equal x (car limits)) *f*)
			((equal x (cdr limits)) *f*)
			(t t))))

	; Return the first safe median of A.
	(find-safe-median (lambda (a)
		(label ((limits (find-limits a))
		(safe-median (lambda (a)
			(cond ((safe-median-p (car a) limits)
					(car a))
				(t (safe-median (cdr a)))))))
		(safe-median a))))

	; Find X mem A for which an Y mem A exists so that P(Y,X)
	; (in other words: find a non-minimum member of A).
	(not-least (lambda (a x)
		(cond ((null a) x)		; wrong, but will be detected by PART
			((p (car a) x) x)
			(t (not-least (cdr a) x)))))

	; Find the middle element of X.
	; Assume RX = (REVERSE X)
	(find-middle (lambda (x rx)
		(cond ((eq (car x) (car rx)) (car x))
			((eq (car x) (cadr rx)) (car x))
			(t (find-middle (cdr x) (cdr rx))))))

	; Test partitions for degenarate cases.
	(degenerate-p (lambda (parts)
		(cond ((null (car parts)) t)
			(t (null (cadr parts))))))

	; Auxilary function used by PART, below.
	(dopart (lambda (a m a1 a2)
		(cond ((null a)
			(list a1 a2))
		((p (car a) m)
			(dopart (cdr a) m (cons (car a) a1) a2))
		(t
			(dopart (cdr a) m a1 (cons (car a) a2))))))

	; Divide the list A into two partitions A1 and A2 where
	;  P(x,MEDIAN) = *T* for each x mem A1
	;  P(x,MEDIAN) = *F* for each x mem A2
	; Return (A1 A2)
	(part (lambda (a)
		(label
			; Use the middle element of A for a default median.
			; Pretty random and easy to find.
			((plist (dopart a (find-middle a (reverse a)) nil nil)))

		(cond ((degenerate-p plist)
				; If the default median is degenerate,
				; try the non-least member of A.
				(label ((nplist (dopart a (not-least a (car a)) nil nil)))
					(cond ((degenerate-p nplist)
							; As a last resort, search a safe median.
							(dopart a (find-safe-median a) nil nil))
				(t nplist))))
			(t plist)))))

	; Quicksort two partitions and append them.
	; Used by QSORT to avoid partitioning the set twice.
	(psort (lambda (plist)
		(append (qsort (car plist)) (qsort (cadr plist)))))

	; Quicksort A. Desired order defined by P.
	(qsort (lambda (a)
		(cond ((null a) nil)	; the empty list is trivial
			((null (cdr a)) a)	; so is the one-element list
			((lognot (cddrp a))	; if there are exactly two members...
				(cond ((p (car a) (cadr a)) a)		; ...swap if necessary
					(t (list (cadr a) (car a)))))
			; if A is ordered, return it
			((orderedp a p) a)
			; otherwise, partition A and PSORT the partitions
			(t (psort (part a)))))))

	(qsort a)))
