; ArrowLISP Example Program
; Copyright (C) 1998-2004 Nils M Holm. All rights reserved.
; See the file LICENSE of the ArrowLISP distribution
; for conditions of use.

; ---purpose---
; Convert arithmetic expressions from prefix to
; infix notation. Prefix expressions may contain
; variables [symbols like X], numbers [quoted
; integer numbers like '#+57], and these functions:
; IPLUS, IDIFFERENCE, ITIMES, IQUOTIENT, IEXPT,
; MINUS. Functions are translated to these
; operators: + - * / ^.
; --
; For instance,
; -
; (INFIX '(IPLUS '#+15 '#+17))
; -
; gives
; -
; '(#15 + #17).
; --
; INFIX will insert parentheses ([ and ]) where
; required.

; ---keywords---
; INFIX function, prefix to infix, RPN, conversion

; ---args---
; X - arithmetic expression

; ---example---
; (infix '(iplus '#+2 (itimes '#+3 (iexpt '#+4 (iplus '#+5 '#+6)))))

:l lib/digits.l
:l lib/caar.l
;l lib/cadr.l
:l lib/caaar.l
;l lib/caadr.l
;l lib/caddr.l
;l lib/cdadr.l
:l lib/caaaar.l
;l lib/caaddr.l
:l lib/append.l
:l lib/member.l
:l lib/assoc.l
:l lib/iplus.l
:l lib/idifference.l
:l lib/itimes.l
:l lib/iquotient.l
:l lib/iexpt.l

; ---code---
(defun infix (x)
	(label

	; Operator symbols
	((+ '+) (- '-) (* '*) (/ '/) (^ '^) ([ '[) (] '])

	; Function->operator translation
	(ops '((iplus.+) (idifference.-) (itimes.*) (iquotient./) (iexpt.^)))

	; Precedence list (descending order)
	(precedence '((high) (iexpt) (itimes iquotient) (iplus idifference) (low)))

	(functionp (lambda (x)
		(member x '(iplus idifference itimes iquotient iexpt))))

	(symbolp (lambda (x)
		(member x '#abcdefghijklmnopqrstuvwxyz)))

	; Numbers are always quoted, like (QUOTE #+5)
	(numericp (lambda (x)
		(cond ((atom x) *f*) (t (eq (car x) 'quote)))))

	; Has the function associated with the operator X
	; a higher precedence than Y?
	(higherprecp (lambda (x y)
		(label ((hp (lambda (x y l)
			(cond ((null l) *f*)
				((member x (car l))
					(cond ((member y (car l)) *f*) (t *t*)))
				((member y (car l)) *f*)
				(t (hp x y (cdr l)))))))
		(hp x y precedence))))

	; Is X the application of a function?
	(operationp (lambda (x)
		(cond ((atom x) *f*)
			((functionp (car x)) *t*)
			(t *f*))))

	(append3 (lambda (a b c) (append a (append b c))))

	(parenthesize (lambda (x) (append3 '#[ x '#])))

	(inf (lambda (x op)
		(cond ((numericp x)
				; Number: remove positive sign
				(cond ((eq (caadr x) -) (cadr x))
					(t (cdadr x))))
			((symbolp x) (list x))
			((eq (car x) 'minus)
				(append (list -) (inf (cadr x) 'high)))
			; Binary function?
			((functionp (car x))
				; Convert prefix to infix
				(label ((sub (append3 (inf (cadr x) (car x))
						(list (cdr (assoc (car x) ops)))
						(inf (caddr x) (car x)))))
				; If the surrounding operation has a higher
				; precedence, parenthesize this expression
				(cond ((higherprecp op (car x))
						(parenthesize sub))
					(t sub))))
			(t (bottom (list 'syntax: x)))))))

	(inf x 'low)))

