; ArrowLISP Example Program
; Copyright (C) 1998-2004 Nils M Holm. All rights reserved.
; See the file LICENSE of the ArrowLISP distribution
; for conditions of use.

; ---purpose---
; Convert arithmetic expressions in infix notation
; to S-expressions. Infix expressions are represented
; by flat lists of variables (atoms) operators (atoms)
; and numbers (sequences of atoms representing digits).
; --
; The following operators are recognized: + (addition)
; - (subtraction), * (multiplication), / (division),
; ^ (exponentation), and - (negation). Brackets are
; recoginzed as parentheses.
; --
; For instance,
; -
; (PREFIX '#57*[X+Y])
; -
; gives
; -
; '(ITIMES (QUOTE #+57) (IPLUS X Y)).

; ---keywords---
; PREFIX function, infix to prefix, RPN, conversion

; ---args---
; X - arithmetic expression

; ---example---
; (prefix '#12+34*56^[7+8])

:l lib/digits.l
:l lib/reverse.l
:l lib/member.l
:l lib/caar.l
;l lib/cadr.l
:l lib/caaar.l
;l lib/caadr.l
;l lib/cdadr.l
:l lib/lognot.l
:l lib/iplus.l
:l lib/idifference.l
:l lib/itimes.l
:l lib/iquotient.l
:l lib/iexpt.l

; ---code---
(defun prefix (x)
	(label

	; Operator symbols
	((+ '+) (- '-) (* '*) (/ '/) (^ '^) ([ '[) (] '])

	(symbolp (lambda (x)
		(member x '#abcdefghijklmnopqrstuvwxyz)))

	(numericp (lambda (x)
		(cond ((atom x) *f*) (t (eq (car x) 'quote)))))

	(operatorp (lambda (x) (member x '#^*/+-)))

	(digitp (lambda (x) (member x digits)))

	; Extract a numeric value from the beginning of X.
	; Return (N XR) where N is the value extracted and
	; XR is the rest of X (X with N removed).
	(number (lambda (x r)
		(cond ((null x)
				(list (list quote (cons + (reverse r))) x))
			((lognot (digitp (car x)))
				(list (list quote (cons + (reverse r))) x))
			(t (number (cdr x) (cons (car x) r))))))

	; These functions are used to extract parts
	; of (EXPR REST) lists where EXPR is the prefix
	; expression built so far and REST is the rest
	; the source expression to parse
	(car-of-rest caadr)
	(cdr-of-rest cdadr)
	(expr car)
	(rest cadr)

	; Parse factors:
	; factor := [ sum ]
	; 	| - factor
	;	| Number
	;	| Symbol
	(factor (lambda (x)
		(cond ((null x) (list x nil))
			; Parse parenthesized subexpressions
			((eq (car x) [)
				(label ((xsub (sum (cdr x))))
				(cond ((null (rest xsub)) (bottom 'missing-paren))
					((eq (car-of-rest xsub) ])
						(list (expr xsub) (cdr-of-rest xsub)))
					(t (bottom 'missing-paren)))))
			; Parse applications of unary minuses
			((eq (car x) -)
				(label ((fac (factor (cdr x))))
				(list (list 'minus (expr fac)) (rest fac))))
			; Parse literal numbers
			((digitp (car x))
				(number x nil))
			; Parse symbols
			(t (list (car x) (cdr x))))))

	; Parse powers:
	; power := factor
	;	| factor ^ power
	(power (lambda (x)
		(label ((left (factor x)))
		(cond ((null (rest left)) left)
			((eq (car-of-rest left) ^)
				(label ((right (power (cdr-of-rest left))))
				(list (list 'iexpt (expr left) (expr right))
					(rest right))))
			(t left)))))

	; Parse terms:
	; term := power
	;	| power Symbol
	;	| power * term
	;	| power / term
	(term (lambda (x)
		(label ((left (power x)))
		(cond ((null (rest left)) left)
			((symbolp (car-of-rest left))
				(label ((right (term (rest left))))
				(list (list 'itimes (expr left) (expr right))
					(rest right))))
			((eq (car-of-rest left) *)
				(label ((right (term (cdr-of-rest left))))
				(list (list 'itimes (expr left) (expr right))
					(rest right))))
			((eq (car-of-rest left) /)
				(label ((right (term (cdr-of-rest left))))
				(list (list 'iquotient (expr left) (expr right))
					(rest right))))
			(t left)))))

	; Parse sums:
	; sum := term
	;	| term + sum
	;	| term - sum
	(sum (lambda (x)
		(label ((left (term x)))
		(cond ((null (rest left)) left)
			((eq (car-of-rest left) +)
				(label ((right (sum (cdr-of-rest left))))
				(list (list 'iplus (expr left) (expr right))
					(rest right))))
			((eq (car-of-rest left) -)
				(label ((right (sum (cdr-of-rest left))))
				(list (list 'idifference (expr left) (expr right))
					(rest right))))
			(t left))))))

	; Pass X to the recursive descent parser consisting of
	; SUM, TERM, POWER, FACTOR. The parsing process returns a
	; list of the form (EXPR REST) as described above. When the
	; REST is NIL, the entire expression could be parsed
	; successfully.
	(label ((px (sum x)))
	(cond ((lognot (null (rest px)))
			(bottom (list 'syntax: (cadr px))))
		(t (expr px))))))

