; ArrowLISP Example Program
; Copyright (C) 1998-2004 Nils M Holm. All rights reserved.
; See the file LICENSE of the ArrowLISP distribution
; for conditions of use.

; ---purpose---
; Evaluate a pure LISP expression in a given environment.
; This function is rather a model than an implementation.
; --
; Primitives supported:
; ATOM BOTTOM CAR CDR COND CONS EQ EXPLODE IMPLODE LABEL
; LAMBDA LIST NULL QUOTE.
; --
; Internally defined symbols: *F* *T* NIL T.

; ---keywords---
; XEVAL function, EVAL, LISP semantics, formal semantics
; evaluate, evaluation

; ---args---
; X - s-expression to evaluate
; E - list of (name.value) pairs.

; ---example---
; ; Evaluate the Towers of Hanoi
; (xeval '(hanoi (quote a) (quote b) (quote c) (quote #iii))
;
;	'((hanoi lambda (a b c n) (cond ((null n) nil)
;		(*t* (append3 (hanoi a c b (cdr n))
;			(cons (cons a (cons b nil)) nil)
;				(hanoi c b a (cdr n))))))
;
;	(append2 lambda (a b)
;		(cond ((null a) b)
;			(*t* (cons (car a) (append2 (cdr a) b)))))
;
;	(append3 lambda (a b c)
;		(append2 a (append2 b c)))))

:l lib/assoc.l
:l lib/pair.l
:l lib/caar.l
;l lib/cadr.l
;l lib/cddr.l
:l lib/caaar.l
;l lib/caadr.l
;l lib/cadar.l
;l lib/caddr.l
:l lib/caaaar.l
;l lib/cadadr.l
;l lib/caddar.l
:l lib/append.l
:l lib/member.l

; ---code---
(defun xeval (x e)
	(label

	; Evaluate each member of list X in environment E:
	; (M1 ... Mn) -> (eval[M1] ... eval[Mn])
	((evlist (lambda (x e)
		(cond ((null x) nil)
			(t (cons (_eval (car x) e) (evlist (cdr x) e))))))

	; Check whether the atom X is a keyword.
	(syntaxp (lambda (x)
		(member x '(bottom quote car cdr cons atom null eq cond
					list implode explode lambda label))))

	; Find the value of a symbol
	(value (lambda (x e)
		(label ((b (assoc x e)))
		(cond ((null b) nil)
			(t (cdr b))))))

	; Create a local environment from a LABEL
	(makeenv (lambda (lbl env)
		(cond ((null lbl) env)
			(t (makeenv (cdr lbl)
				(cons (cons (caar lbl) (_eval (cadar lbl) env)) env))))))

	(_eval (lambda (x e)
		; --- Atoms ---
		(cond ((atom x)
				(cond ((null x) nil)	; empty list
					((eq x *t*) *t*)	; predefined symbols
					((eq x *f*) *f*)
					((eq x 'nil) nil)
					((eq x 't) *t*)
					(t (value x e))))	; default: retrieve binding from E
			; --- Lists with atom at first slot ---
			((atom (car x))
				; first try the literal keywords
				(cond ((eq (car x) 'bottom)
						(bottom (cadr x)))
					((eq (car x) 'quote)
						(cadr x))
					((eq (car x) 'car)
						(car (_eval (cadr x) e)))
					((eq (car x) 'cdr)
						(cdr (_eval (cadr x) e)))
					((eq (car x) 'cons)
						(cons (_eval (cadr x) e) (_eval (caddr x) e)))
					((eq (car x) 'atom)
						(atom (_eval (cadr x) e)))
					((eq (car x) 'null)
						(null (_eval (cadr x) e)))
					((eq (car x) 'eq)
						(eq (_eval (cadr x) e) (_eval (caddr x) e)))
					((eq (car x) 'cond)
						(cond ((eq (_eval (caadr x) e) *t*)
								(_eval (cadadr x) e))
							((eq (_eval (caadr x) e) *f*)
								(_eval (cons 'cond (cddr x)) e))
							(t (bottom 'xeval:cond))))
					((eq (car x) 'list)
						(evlist (cdr x) e))
					((eq (car x) 'implode)
						(implode (_eval (cadr x) e)))
					((eq (car x) 'explode)
						(explode (_eval (cadr x) e)))
					; lambda functions already are in their normal forms
					((eq (car x) 'lambda)
						x)
					((eq (car x) 'label)
						(_eval (caddr x) (makeenv (cadr x) e)))
					; At this point, the first member must be a variable.
					; If it is bound to a keyword, evaluate it and try
					; again.
					((syntaxp (_eval (car x) e))
						(_eval (cons (_eval (car x) e) (cdr x)) e))
					; If it is bound to a non-keyword atom, this is an
					; application of a non-function
					((atom (_eval (car x) e))
						(bottom 'xeval:non-function))
					; By default, replace the variable with its binding
					; and try again
					(t
						(_eval (cons (_eval (car x) e) (cdr x)) e))))
			; --- Applications of lambda functions ---
			((eq (caar x) 'lambda)
				; Bind the actual arguments to the formal arguments
				; and then evaluate the term of the lambda function
				; in the union of the new bindings and E
				(_eval (caddar x)
					(append (pair (cadar x) (evlist (cdr x) e)) e)))
			; --- Default: undefined ---
			(t (bottom 'xeval))))))

	(_eval x e)))
